##################################################################################
# Copyright (c) 2011, Daniel Urieli, Peter Stone
# University of Texas at Austin
# All right reserved
# 
# Based On:
# 
# Copyright (c) 2000-2003, Jelle Kok, University of Amsterdam
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# 
# 3. Neither the name of the University of Amsterdam nor the names of its
# contributors may be used to endorse or promote products derived from this
# software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
##################################################################################




import sys
import os
import random
from math import *
from numpy import *

import seaModels
import shipModels
import state
import agents
import simulationEnvironment
import factories

#from heuristicDivide import *
from geometry import *

def default(str):
  return str + ' [Default: %default]'

def validateOptionCombinations(args):
  """
  This functions validates we didn't mistakenly chosen options
  that don't work with each other.
  If options are verified we just return to the the program
  execution. If the options are incorrect, we gracefully exit.
  """
#  if( args['shipType'] == 'basic' 
#      and args['worldModel'] == 'complete' 
#      and args['strategy'] == 'circling' ):
#     return

#  if( args['shipType'] == 'basic'
#      and args['worldModel'] == 'complete'
#      and args['strategy'] == 'holonomicrrt' ):
#     return

#  if( args['shipType'] == 'basic'
#      and args['worldModel'] == 'complete'
#      and args['strategy'] == 'pid' ):
#     return

#  if( args['shipType'] == 'basic'
#      and args['worldModel'] == 'complete'
#      and args['strategy'] == 'measuredist' ):
#     print 'Warning: change this combination in the future'
#     return

  if( args['shipType'] == 'basic'
      and args['worldModel'] == 'complete'
      and args['strategy'] == 'basicpatrol' ):
     print 'Warning: change this combination in the future'
     return

  if( args['shipType'] == 'basic'
      and args['worldModel'] == 'complete'
      and args['strategy'] == 'rrt' ):
     print 'Warning: change this combination in the future'
     return

  print 'Illegal options combination - exiting...'
  sys.exit()

  # TODO: Once we support mixed ship types, check combinations 
  #       per each ship
  



def readCommand( argv ):
  """
  Processes the command used to run the simulation from the command line.
  """
  from optparse import OptionParser
  usageStr = """
  USAGE:      python main.py <options>
  EXAMPLE:    python main.py --option <value> # TODO complete with real values
                - #TODO explain what the example command do
  """
  parser = OptionParser(usageStr)
  
#  parser.add_option('-n', '--numGames', dest='numGames', type='int',
#                    help=default('the number of GAMES to play'), metavar='GAMES', default=1)
#  parser.add_option('-l', '--layout', dest='layout', 
#                    help=default('the LAYOUT_FILE from which to load the map layout'), 
#                    metavar='LAYOUT_FILE', default='oneHunt')
#  parser.add_option('-p', '--pacman', dest='pacman', 
#                    help=default('the agent TYPE in the pacmanAgents module to use'), 
#                    metavar='TYPE', default='BustersKeyboardAgent')
#  parser.add_option('-a','--agentArgs',dest='agentArgs',
#                    help='Comma seperated values sent to agent. e.g. "opt1=val1,opt2,opt3=val3"')
#  parser.add_option('-g', '--ghosts', dest='ghost', 
#                    help=default('the ghost agent TYPE in the ghostAgents module to use'), 
#                    metavar = 'TYPE', default='RandomGhost')

  parser.add_option('-q', '--quietTextGraphics', action='store_true', dest='quietGraphics', 
                    help=default('Generate minimal output and no graphics'), default=False )    
  parser.add_option('-k', '--numSteps', dest='numSteps', type='int',
                    help=default("Number of steps to simulate"), default=10000)
  parser.add_option('-d', '--inputFilesDir', dest='inputFilesDir', 
                    help=default('the DIR in which input files are searched for'), 
                    metavar='DIR', default='input_files')
  parser.add_option('-f', '--patrolFile', dest='patrolFile', 
                    help=default('A file defining patrol-points, ship initial positions, and patrol paths. The file is assumed to be under the input_files directory'), 
                    default='shipPathsAndPositions.py')
  parser.add_option('-r', '--resultsType', dest='resultsType', 
                    help=default('Type of results the simulator outputs.'), 
                    default='edgeGraphData')
  parser.add_option('-t', '--shipType', dest='shipType',
                    help=default("The ship's model"), default='basic')
  parser.add_option('-w', '--worldModel', dest='worldModel',
                    help=default("World model that the agent maintains"), default='complete')
  parser.add_option('-s', '--strategy', dest='strategy',
                    help=default("Agent's desicion making strategy"), default='basicpatrol')
  parser.add_option('-y', '--rulesOfTheSea', action='store_true', dest='rulesOfTheSea', 
                    help=default('Respect the rules of the sea and yield when needed'), default=False )    
#### numShips is disabled because determined by the number of paths
#  parser.add_option('-n', '--numShips', dest='numShips', type='int',
#                    help=default("Number of ships"), default=1)


#  parser.add_option('-k', '--numghosts', type='int', dest='numGhosts', 
#                    help=default('The maximum number of ghosts to use'), default=4)
#  parser.add_option('-z', '--zoom', type='float', dest='zoom', 
#                    help=default('Zoom the size of the graphics window'), default=1.0)
#  parser.add_option('-f', '--fixRandomSeed', action='store_true', dest='fixRandomSeed', 
#                    help='Fixes the random seed to always play the same game', default=False)
#  parser.add_option('-s', '--showGhosts', action='store_true', dest='showGhosts', 
#                    help='Renders the ghosts in the display (cheating)', default=False)
#  parser.add_option('-t', '--frameTime', dest='frameTime', type='float',
#                    help=default('Time to delay between frames; <0 means keyboard'), default=0.1)
  
  options, other = parser.parse_args()
  if len(other) != 0: 
    raise Exception('Command line input not understood: ' + other)
  args = dict()

#  # Fix the random seed
#  if options.fixRandomSeed: random.seed('bustersPacman')
#  
#  # Choose a layout
#  args['layout'] = layout.getLayout( options.layout )
#  if args['layout'] == None: raise Exception("The layout " + options.layout + " cannot be found")
#  
#  # Choose a ghost agent
#  ghostType = loadAgent(options.ghost, options.quietGraphics)
#  args['ghosts'] = [ghostType( i+1 ) for i in range( options.numGhosts )]
#
#  # Choose a Pacman agent
#  noKeyboard = options.quietGraphics
#  pacmanType = loadAgent(options.pacman, noKeyboard)
#  agentOpts = parseAgentArgs(options.agentArgs)
#  agentOpts['ghostAgents'] = args['ghosts']
#  pacman = pacmanType(**agentOpts) # Instantiate Pacman with agentArgs
#  args['pacman'] = pacman
#        
#  import graphicsDisplay
#  args['display'] = graphicsDisplay.FirstPersonPacmanGraphics(options.zoom, \
#                                                                options.showGhosts, \
#                                                                frameTime = options.frameTime)
#  args['numGames'] = options.numGames
  
  # Choose GUI / non-GUI
  if options.quietGraphics:
    args['withDisplay'] = False
  else:
    args['withDisplay'] = True
    
  args['shipType'] = options.shipType
  args['worldModel'] = options.worldModel
  args['strategy'] = options.strategy

#  args['numShips'] = options.numShips

  args['numSteps'] = options.numSteps

  args['inputFilesDir'] = options.inputFilesDir

  args['patrolFile'] = options.patrolFile

  args['resultsType'] = options.resultsType

  args['rulesOfTheSea'] = options.rulesOfTheSea

  validateOptionCombinations(args)
  
  return args

######################################################
# Initialization Functions -                         #
# Those function are here temporarily                # 
# until there is a flexible solution                 #
# for loading different models, perhaps              #
# as a part of the task definition language          #
######################################################
def initWindFromFile(filePath):
  # just execute the (python) file and return a variable generated
  exec(compile(open(filePath).read(), filePath, 'exec'))
  return wind

def initWaterCurrentsFromFile(filePath):
  # just execute the (python) file and return a variable generated
  exec(compile(open(filePath).read(), filePath, 'exec'))
  return waterCurrents

def initWavesFromFile(filePath):
  # just execute the (python) file and return a variable generated
  exec(compile(open(filePath).read(), filePath, 'exec'))
  return waves

def initObstaclesFromFile(filePath):
  # just execute the (python) file and return a variable generated
  exec(compile(open(filePath).read(), filePath, 'exec'))
  return obstacles

def initShipPatrolPointsAndPathsFromFile(filePath):
  # just execute the (python) file and return variables generated
  exec(compile(open(filePath).read(), filePath, 'exec'))
  return paths, patrolPoints, shipExternalStates

######################################################
# End initialization Functions                       #
######################################################



######################################################
# Results writing functions                          #
######################################################

def writeResults(resultsType, env, patrolPoints):
  """
  Write the simulation results.

  TODO: this function needs to be further organized
  once we have a better picture regarding what
  types of results are usually needed.
  Also the function arguments should be revised as a part
  of the function reorganization.

  @type  resultsType: string
  @param resultsType: a string identifier for the type of result to write

  @type  resultsType: SimulationEnvironment
  @param resultsType: The simulation environment, used to extract data from

  @type  patrolPoints: array of (x,y) tuples
  @param patrolPoints: All the points that participate in the patrol. This is required when writing data for heuristicDivide algorithm, and might be removed at some point.
  """
  #TODO: this function needs to be further organized
  #      once we have a better picture regarding what
  #      types of results are usually needed.
  if resultsType == 'worstFreq':

    f = open(resultsType + '.res', 'w')

    # Simulation results, K+1, K-1
    point2freq = dict([(point, mean([l[i] - l[i-1] for i in range(1,len(l))])) for point, l in env.visitTimes.items() ])
    worstCaseFreq1 = max(point2freq.values())
#    print >> sys.stderr, 'WORSTCASE FREQ:', worstCaseFreq1 
    f.write('WORSTCASE FREQ: ' + str(worstCaseFreq1) + '\n')
    f.write('patrolPoints=' + str(patrolPoints) + '\n')
    f.write('visitTimes=' + str(env.visitTimes) + '\n')
#    print >>sys.stderr, 'visitTimes =', env.visitTimes

#    print >> sys.stderr, 'numShipsForPatrol', numShips
#    worstCaseFreq2 = max( [mean([l[i] - l[i-1] for i in range(1,len(l))]) for l in env.state.times.values()] )
#    print >> sys.stderr, 'verification WORSTCASE FREQ:', worstCaseFreq2

#    if worstCaseFreq1 != worstCaseFreq2:
#      raise Exception('Whats going on? different worst cases freqs')

    f.close()

  elif resultsType == 'edgeGraphData':

    f = open(resultsType + '.py', 'w')
    f.write('patrolPoints=' + str(patrolPoints) + '\n')
    f.write('edgeLengths=' + str(env.edgeLengths) + '\n')
    f.close()

######################################################
# End Results writing functions                      #
######################################################


    





#def run(withDisplay, shipType, worldModel, strategy, numShips, inputFilesDir, patrolFile):
# disabled numShips, because determined by the number of paths
def run(withDisplay, shipType, worldModel, strategy, inputFilesDir, patrolFile, numSteps, resultsType, rulesOfTheSea):
  """
  The main flow of the application. The arguments are generated by command line options, or their defaults.

  @type  withDisplay: boolean
  @param withDisplay: Tells whether to run in a GUI mode or in non-GUI mode

  @type  shipType: string
  @param shipType: The ship model to allocate

  @type  worldModel: string
  @param worldModel: The world model the agent has

  @type  strategy: string
  @param strategy: The decision-making strategy the agent has

  @type  inputFilesDir: string
  @param inputFilesDir: The directory in which the input files are 

  @type  patrolFile: string
  @param patrolFile: A file that defines ship positions, ship paths, and the patrol points of interest

  @type  numSteps: int
  @param numSteps: Number of steps to simulate

  @type  resultsType: string
  @param resultsType: Chooses which results to write 

  @type  rulesOfTheSea: boolean
  @param rulesOfTheSea: Tells whether to respect the rules of the sea
  """
 
  wind = initWindFromFile(os.path.join(inputFilesDir, 'wind.py'))
  waterCurrents = initWaterCurrentsFromFile(os.path.join(inputFilesDir, 'waterCurrents.py'))
  waves = initWavesFromFile(os.path.join(inputFilesDir, 'waves.py'))
  obstacles = initObstaclesFromFile(os.path.join(inputFilesDir, 'obstacles.py'))
  sea = seaModels.Sea(wind, waterCurrents, waves, obstacles)

  paths, patrolPoints, shipExternalStates = initShipPatrolPointsAndPathsFromFile(os.path.join(inputFilesDir, patrolFile))
#  print 'paths', paths
#  print 'patrolPoints', patrolPoints
#  print 'shipExternalStates', shipExternalStates

  # used to be sent as a flag, but doesn't seem like there is a need
  numShips = len(paths)

  # Choose a ship factory
  if shipType == 'basic':
    shipFactory = factories.BasicShipFactory()
  else:
    raise Exception('Unknown ship type: ' + shipType)

  # Allocate ships  
  ships = [shipFactory.create() for i in range(numShips)]

  # world state
  initialState = state.State( sea, ships, shipExternalStates )


  # Allocate an agent for each ship
  # TODO: When we support mixed ship models,
  #       allocate an agent based on ship, and send it the possible actions 
  #       for this ship

  # Choose a world model factory
  if worldModel == 'complete':
    worldModelFactory = factories.CompleteWorldModelFactory()
  else:
    raise Exception('Unknown world model type: ' + worldModel)

  # Choose a decision making strategy
  if strategy == 'circling':
    strategyFactory = factories.CirclingStrategyFactory()
  elif strategy == 'rrt':
    strategyFactory = factories.RRTStrategyFactory(paths, rulesOfTheSea)
  elif strategy == 'pid':
    strategyFactory = factories.PIDStrategyFactory(rulesOfTheSea)
  elif strategy == 'measuredist':
    strategyFactory = factories.MeasureEdgeLengthsFactory()
  elif strategy == 'basicpatrol':
    strategyFactory = factories.AgentBasicPatrolStrategyFactory(paths, rulesOfTheSea)
  else:
    raise Exception('Unknown decision making strategy: ' + strategy)

  # Allocate agents with the above factories
  shipAgents = [agents.Agent( worldModel = worldModelFactory.create(shipIndex=i),
                              strategy = strategyFactory.create(agentIndex=i) ) 
                              for i,s in enumerate(ships)]


  # Simulation module
  env = simulationEnvironment.SimulationEnvironment( initialState, shipAgents )


  # start the simulation
  import mainGUI
  if withDisplay:
    mainGUI.runGui(env, shipType, worldModel, strategy, numSteps)
  else:
    env.run(numSteps)


  # write results
  writeResults(resultsType, env, patrolPoints)




if __name__ == '__main__':
  """
  The main function called when running the simulation

  > python main.py

  See the usage string for more details.

  > python main.py --help
  """
  args = readCommand( sys.argv[1:] ) # Get game components based on input
  print 'Running simulation with the following options:'
  for k, v in sorted(args.items()):
    print k, '=', v
  run( **args )
